/* 
 * Description: ralf: Rename A Lot of Files
 * Version:  0.3
 * Licence:  MIT
 * Author:  prx prx@si3t.ch
 */



#include <err.h>
#include <limits.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/queue.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "ralf.h"

size_t
estrlcat(char *dst, const char *src, size_t dstsize)
{
	size_t size = 0;


	if ((size = strlcat(dst, src, dstsize)) >= dstsize)
		err(1, "strlcat on %s + %s", dst, src);

	return size;
}

size_t
estrlcpy(char *dst, const char *src, size_t dstsize)
{
	size_t n = 0;


	n = strlcpy(dst, src, dstsize);
	if (n >= dstsize)
		err(1, "strlcpy failed for %s = %s", dst, src);

	return n;
}

int
file_exist(const char *fn)
{
	struct stat sb;


	if (stat(fn, &sb) == 0)
		return 1;
	else
		return 0;
}

void
get_input(char *input, size_t strsiz)
{
	size_t inputsize = 0;
	char *tmp        = NULL;


	fflush(stdin);
	if (getline(&tmp, &inputsize, stdin) == -1)
		if (ferror(stdin))
			errx(1, "An error occured while reading you search");

	/* remove newline */
	tmp[strcspn(tmp, "\n")] = '\0';

	estrlcpy(input, tmp, strsiz);

	free(tmp);
}

/* sere : search n replace */
int
sere(const char *str, const char *pattern, const char *rep,
	              char *new, size_t newsiz)
{
	/*
	 * search pattern in str
	 * replace with rep as long as pattern exists
	 * record the new string in new
	 * return how many replacement were made
	 * if empty name, return -1
	 */

	int regret             = 0;
	int n                  = 0;
	int lastpos            = 0;
	char buf[BUFSIZ]       = {'\0'};
	char tmp[FILENAME_MAX] = {'\0'};
	regex_t regex;
	regmatch_t match[1];

	/* copy the string to tmp, then modify it */
	estrlcpy(tmp, str, sizeof(tmp));

	if ((regret = regcomp(&regex, pattern, REG_EXTENDED)) != 0) {
		regerror(regret, &regex, buf, sizeof(buf));
		regfree(&regex);
		err(1, "%s", buf);
	}

	while (regexec(&regex, tmp, 1, match, 0) == 0) {

		/* append beginning of tmp in new until match start */

		/* stop string before match */
		tmp[match[0].rm_so] = '\0';

		/* copy the beginning before match */
		estrlcat(new, tmp, newsiz);

		/* replace this pattern part */
		if (strlen(rep) > 0)
			estrlcat(new, rep, newsiz);

		/* move the end of the string at the beginning of tmp */
		memmove(tmp, tmp + match[0].rm_eo, sizeof(tmp) - match[0].rm_eo);

		n++;
	}
	/* at the end, copy the unmatching ending of tmp in new */ 
	estrlcat(new, tmp, newsiz);

	regfree(&regex);


	return n;
}

void
usage(void)
{
        errx(1, "usage: %s [file ...]\n", getprogname());
}

int 
main(int argc, char *argv[])
{
	int longest_name                  = 0;
	int to_rename                     = 0;
	char *slash	                  = NULL;
	char pattern[FILENAME_MAX] 	  = {'\0'};
	char replace[FILENAME_MAX]        = {'\0'};
	char **files	                  = NULL;
	struct Filelist_head filelist_head;
	struct Filename *fn, *fnp;

#ifdef __OpenBSD__
	if (pledge("stdio rpath cpath", NULL) == -1)
		err(1, "pledge failed");
#endif


	if (!(argc > 1))
		usage();

	SIMPLEQ_INIT(&filelist_head);

	/* start at i = 1 since we don't want to rename self */
	for (int i=1; i < argc; i++) {
		/* make sure the file exists, or skip it */
		if (!(file_exist(argv[i]))) {
			warnx("\"%s\" not found, skip...", argv[i]);

		} else {
			if ((fn = malloc(sizeof(struct Filename))) == NULL )
				err(1, "malloc");
			/* initialize structure */
			memset(fn, 0, sizeof(struct Filename));

			/* equivalent to below
			* fn->modified = 0;
			* bzero(fn->name, sizeof(fn->name));
			* bzero(fn->new, sizeof(fn->new));
			*/

			estrlcpy(fn->name, argv[i], sizeof(fn->name));
			SIMPLEQ_INSERT_TAIL(&filelist_head, fn, filename);
		}
	}

	/* display files */
	SIMPLEQ_FOREACH(fnp, &filelist_head, filename) {
		printf("%s\n", fnp->name);
		if (strlen(fnp->name) > longest_name)
			longest_name = strlen(fnp->name);
	}
	printf("%s\n", separator);

	/* get user input */
	/* pattern to search */
	while (strlen(pattern) == 0) {
		printf("%s: ", pattern_input);
		get_input(pattern, sizeof(pattern));
	}

	/* check input */
	while (!(strlen(replace) > 0) || (replace[0] == '\0') ||
		((slash = strchr(replace, '/')) != NULL)) 
	{
		if (slash != NULL)
			warnx("wrong filename, prefer using only [a-zA-Z0-9]");
		printf("%s: ", replace_input);
		get_input(replace, sizeof(replace));
	}

	/* compute new filenames and display changes */
	printf("%s\n", separator);
	SIMPLEQ_FOREACH(fnp, &filelist_head, filename) {
		fnp->modified = sere(fnp->name, pattern, replace,
				     fnp->new, sizeof(fnp->new));
		to_rename += fnp->modified;

		/* filename with appropriate number of spaces to align */
		printf("%s%*s -> ", fnp->name,
				    longest_name - (int)strlen(fnp->name), "");
		if (fnp->modified > 0)
			printf("%s%s%s\n", hi_clr, fnp->new, hi_reset);
		else if (fnp->modified < 0)
			printf("%s%s %s %s\n", hi_clr, fnp->new,
			       invalid_msg, hi_reset);
		else
			printf("%s\n", fnp->new);
	}

	/* ask for confirmation */
	printf("%s\n", separator);
	printf("%s\n", confirm_input);
	fflush(stdin);
	if (getchar() == '\n') {
		/* rename for real now */
		SIMPLEQ_FOREACH(fnp, &filelist_head, filename) {
			if (fnp->modified == 0)
				continue;
			if (file_exist(fnp->new)) {
				warnx("%s already exist, %s is left unchanged",
				       fnp->new, fnp->name);
				continue;
			}

			/* actually rename the file */
			if (rename(fnp->name, fnp->new) == 0) {
				printf("%s%*s -> %s\n", fnp->name,
				    longest_name - (int)strlen(fnp->name), "",
				    fnp->new);
			} else {
				warn("Error while renaming %s to %s",
				       fnp->name, fnp->new);
			}
		}

	}
	printf("bye o/\n");

	free(fn);

	return 0;
}
